# Copyright (c) 2021, 2022 Yoshihiro Ota <ota@j.email.ne.jp>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import os
import logging
from typing import Dict

logger = logging.getLogger(__name__)


def _clone_with_primitives(input: Dict) -> Dict:
    return {k: v for k, v in input.items() if type(v) in [str, int, float]}


class Incident:
    levels = ["error", "warn", "info"]

    class Level:
        def __init__(self, level: str, params: Dict, fallback: Dict = {}):
            assert level
            self._level = level
            self._vars: Dict = _clone_with_primitives(params)
            self._vars["level"] = level
            for key in ["trigger", "untrigger", "escalation"]:
                if key not in params:
                    if key in fallback:
                        self._vars[key] = fallback[key]
                    else:
                        raise Exception(f"'{key}' is missing in '{level}' level")

            self._trigger: str = self._vars["trigger"]
            self._untrigger: str = self._vars["untrigger"]
            self._escalation: str = self._vars["escalation"]

            self._triggered: bool = False

        def clear(self) -> None:
            self._triggered = False

        def escalate_if_in_range(self, locals: Dict) -> bool:
            my_locals = {**locals, **self._vars}
            logger.debug(f"Checking trigger='{self._trigger}' at level={self._level}")
            expr = eval(f'f"{self._trigger}"', {"__builtins__": {}}, my_locals)
            logger.debug(f"Resolved to expression='{expr}'")
            if eval(expr, {"__builtins__": {}}, my_locals):
                if not self._triggered:
                    cmd = eval(f'f"{self._escalation}"', my_locals)
                    logger.debug(f"Escalating at level={self._level} with cmd=[{cmd}]")
                    self._triggered = True
                    os.system(cmd)
                return True
            if self._triggered:
                logger.debug(
                    f"Checking untrigger='{self._untrigger}' at level='{self._level}"
                )
                expr = eval(f'f"{self._untrigger}"', {"__builtins__": {}}, my_locals)
                logger.debug(f"Resolved to expression='{expr}'")
                if eval(expr, {"__builtins__": {}}, my_locals):
                    self._triggered = False
                    logger.debug(f"Untriggered at level={self._level}")
                else:
                    return True
            return False

        def verify(self, locals: Dict) -> None:
            my_locals = {**locals, **self._vars}
            expr = eval(f'f"{self._trigger}"', {"__builtins__": {}}, my_locals)
            logger.debug(f"Resolved '{self._trigger}' to trigger='{expr}'")
            expr = eval(f'f"{self._untrigger}"', {"__builtins__": {}}, my_locals)
            logger.debug(f"Resolved '{self._untrigger}' to untrigger='{expr}'")
            expr = eval(f'f"{self._escalation}"', my_locals)
            logger.debug(f"Resolved '{self._escalation}' to escalation='{expr}'")

    def __init__(self, name: str, params: Dict):
        if "description" in params:
            self._desc = params["description"]
        else:
            raise Exception("'description' is missing")

        count = 0
        self._levels = {}
        fallback = {}
        err = None
        for key in Incident.levels:
            try:
                level = Incident.Level(key, params[key], fallback)
                self._levels[key] = level
                count += 1
                fallback = level._vars
            except Exception as e:
                if err is None:
                    err = e
                self._levels[key] = None
        if count == 0:
            if err:
                raise Exception(f"{err} of '{name}' incident")
            l = [f"'{i}'" for i in Incident.levels]
            msg = f"One or more of {', '.join(l[:-1])} and/or {l[-1]} must be specified"
            raise Exception(msg)
        self._name: str = name
        self._vars: Dict = _clone_with_primitives(params)

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other) -> bool:
        return self._name == other._name

    @property
    def name(self) -> str:
        return self._name

    def escalated(self, locals: Dict) -> bool:
        in_range = False
        my_locals = None
        for key in Incident.levels:  # be explicit about ordering
            level = self._levels[key]
            if level:
                if in_range:
                    level.clear()
                else:
                    if my_locals is None:
                        my_locals = {**locals, **self._vars}
                    if level.escalate_if_in_range(my_locals):
                        in_range = True
        return in_range

    def verify(self, locals: Dict) -> None:
        for key, level in self._levels.items():
            if level:
                level.verify({**locals, **self._vars})



# Copyright (c) 2021, 2022 Yoshihiro Ota <ota@j.email.ne.jp>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import copy
import logging
import signal
import threading
from typing import Dict, Optional, Set, Tuple

from . import Incident, instantiate_variable, Variable

logger = logging.getLogger(__name__)


class Monitor:
    _functions = {"__builtins__": {"abs": abs, "len": len, "max": max, "min": min}}

    def __init__(self, interval: float = -1):
        self._interval = interval
        self._constants: Set[Variable] = set()
        self._variables: Set[Variable] = set()
        self._derivatives: Dict[str, str] = {}
        self._incidents: Set[Incident] = set()
        self._locals: Dict[str, Any] = {}
        self._running: Optional[Event] = None

        if self._interval > 0:
            signal.signal(signal.SIGINT, self.exit)
            signal.signal(signal.SIGTERM, self.exit)
            logger.info(f"Monitoring is set for {interval} seconds.")

    def exit(self, *args):
        logger.info(f"Exiting")
        if self._running is not None:
            self._running.set()

    def load_json(self, json: Dict) -> Tuple[int, int, int, int]:
        constants = 0
        variables = 0
        derivatives = 0
        incidents = 0
        if "constants" in json:
            logger.debug(f"Loading constants")
            constants = self.add_constants(json["constants"])
        if "variables" in json:
            logger.debug(f"Loading variables")
            variables = self.add_variables(json["variables"])
        if "derivatives" in json:
            logger.debug(f"Loading derivatives")
            derivatives = self.add_derivatives(json["derivatives"])
        if "incidents" in json:
            logger.debug(f"Loading incidents")
            incidents = self.add_incidents(json["incidents"])
        return (constants, variables, derivatives, incidents)

    def add_constants(self, json: Dict) -> int:
        return self._parse_variables(json, self._constants)

    def add_variables(self, json: Dict) -> int:
        return self._parse_variables(json, self._variables)

    def add_derivatives(self, json: Dict) -> int:
        count = 0
        for key, value in json.items():
            for container in [self._constants, self._variables, self._derivatives]:
                if key in container:
                    raise Exception(f"Variable '{key}' already exists")

            self._derivatives[key] = value
            logger.info(f"Derivative '{key}' is configured")
            count += 1
        return count

    def add_incidents(self, json: Dict) -> int:
        count = 0
        for key, json in json.items():
            incident = Incident(key, json)
            if incident in self._incidents:
                raise Exception(f"Incident '{key}' already exists")
            self._incidents.add(incident)
            logger.info(f"Incident '{incident.name}' is configured")
            count += 1
        return count

    def start(self) -> None:
        self.fetch_constants()

        self.fetch_and_evaluate()
        if self._interval > 0:
            thread: threading.Thread = None
            self._running = threading.Event()
            while not self._running.wait(self._interval):
                if thread:
                    thread.join()
                if not self._running.is_set():
                    thread = threading.Thread(target=self.fetch_and_evaluate)
                    thread.start()

    def fetch_and_evaluate(self) -> None:
        locals = copy.deepcopy(self._locals)
        try:
            self.fetch_variables(locals)
            self.evaludate_derivatives(locals)
            self.evaluate_incidents(locals)
        except IndexError:
            pass

    def fetch_constants(self) -> None:
        for v in self._constants:
            self._locals[v.name] = v.value
            logger.info(f"Constant '{v.name}' holds {v.value}")

    def fetch_variables(self, locals: Dict) -> None:
        for v in self._variables:
            v.new_value()
            locals[v.name] = v
            logger.info(f"'{v.name}' is loaded and holds {v}")
        logger.debug("Reloaded all variables")

    def evaludate_derivatives(self, locals: Dict) -> None:
        for v, expr in self._derivatives.items():
            expr = eval(f'f"{expr}"', Monitor._functions, locals)
            logger.debug(f"Resolved derivative={v} to expression='{expr}'")
            value = eval(expr, Monitor._functions, locals)
            locals[v] = value
            logger.info(f"'{v}' is calculated and holds '{value}'")
        logger.debug("Calculated all derivatives")

    def evaluate_incidents(self, locals: Dict) -> None:
        for incident in self._incidents:
            logger.debug(f"Evaluating '{incident.name}' incident")
            incident.escalated(locals)
        logger.debug("Evaluated all incidents")

    def _parse_variables(self, json: Dict, container: Set) -> int:
        count = 0
        for key, json in json.items():
            variable = instantiate_variable(key, json)
            for existing in [self._constants, self._variables, self._derivatives]:
                if variable in existing:
                    raise Exception(f"Variable '{key}' already exists")
            container.add(variable)
            logger.info(f"Variable '{variable.name}' is configured")
            count += 1
        return count

    def verify(self) -> None:
        self.fetch_constants()
        self.fetch_variables(self._locals)
        self.evaludate_derivatives(self._locals)

        for incident in self._incidents:
            incident.verify(self._locals)



# Copyright (c) 2021, 2022 Yoshihiro Ota <ota@j.email.ne.jp>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import argparse
import logging
import json
import signal
import sys
import os

from . import Monitor, Incident

logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=argparse.FileType("r"),
        required=True,
        nargs="+",
        help="configuration file to specify sysctl monitoring and their actions",
    )

    parser.add_argument(
        "-i",
        "--interval",
        dest="interval",
        type=float,
        default=-1,
        help="positive number to specify interval in second to re-evaluate rules",
    )

    parser.add_argument(
        "-l",
        "--log",
        dest="log",
        type=str,
        required=False,
        help="the name of the log file.  If not specified, logging is disabled",
    )

    parser.add_argument(
        "--log-format",
        dest="logformat",
        type=str,
        default="%(asctime)s %(levelname)s %(name)s | %(message)s",
        help="log line format",
    )

    parser.add_argument(
        "--log-dateformat",
        dest="logdateformat",
        type=str,
        default="%Y-%m-%d_%H:%M:%S",
        help="log date format",
    )

    parser.add_argument(
        "--levels",
        dest="levels",
        type=str,
        default=["error", "warn", "info"],
        nargs="+",
        help="specify custom levels",
    )

    parser.add_argument(
        "--verify",
        dest="verify",
        action="store_true",
        help="semi-dry run to verify input configuration - fetch and evaluate all but do not execute escalation. SyntaxError indicates error in expressions, NameError indicates wrong variable name",
    )
    parser.set_defaults(verify=False)

    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        help="use debug level logging",
    )
    parser.set_defaults(debug=False)

    return parser.parse_args()


def analyze(args) -> None:
    if args.verify:
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.DEBUG,
            format="%(message)s",
            datefmt=args.logdateformat,
        )
    elif args.log:
        logging.basicConfig(
            filename=args.log,
            level=logging.DEBUG if args.debug else logging.INFO,
            format=args.logformat,
            datefmt=args.logdateformat,
        )
    else:
        logging.disable(logging.CRITICAL)

    Incident.levels = args.levels
    m = Monitor(args.interval)
    for file in args.config:
        with file as json_file:
            setting = json.load(json_file)
            counts = m.load_json(setting)
            logger.info(f"Loaded from '{file.name}'")
            logger.info(f"Loaded {counts[0]} constants")
            logger.info(f"Loaded {counts[1]} variables")
            logger.info(f"Loaded {counts[2]} derivatives")
            logger.info(f"Loaded {counts[3]} incidents")
    if args.verify:
        m.verify()
    else:
        m.start()


def main():
    args = parse_args()
    try:
        analyze(args)
    except KeyboardInterrupt:
        try:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            sys.exit(0)
        except SystemExit:
            os._exit(0)


if __name__ == "__main__":
    main()



# Copyright (c) 2021, 2022 Yoshihiro Ota <ota@j.email.ne.jp>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import logging
import os
import struct
import sys
import time
from abc import ABC, abstractmethod
from typing import Any, Dict

from .libc import sysctl

logger = logging.getLogger(__name__)


class Variable(ABC):
    """
    A user defines a variable with its "name" and a way to fetch its value.
    """

    def __init__(self, name: str, typename: str, params: Dict):
        assert name
        if params.get("type", None) != typename:
            raise TypeError(f"Not {typename} type")
        if params.get(typename, None) is None:
            raise TypeError(f"Incomplete {typename} type specification")
        self._name = name
        self._value = None
        self._depth = params.get("depth", None) or params.get("history", 0)
        if self._depth < 0:
            self._depth = 0
        if self._depth != 0:
            self._hist = list()
        else:
            self._hist = None

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other) -> bool:
        return self._name == other._name

    def __repr__(self) -> Any:
        return str(self.value)

    def __getitem__(self, key) -> Any:
        # if hasattr(self, '__dict__'):
        if self._hist is not None:
            try:
                return self._hist[key]
            except IndexError:
                raise
        else:
            return self.value[key]

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> Any:
        return self._value

    @property
    def last_value(self) -> Any:
        if len(self._hist) > 0:
            return self._hist[-1]
        return None

    @property
    def oldest_value(self) -> Any:
        if len(self._hist) > 0:
            return self._hist[0]
        return None

    def new_value(self) -> Any:
        if self._value is not None:
            if self._hist is not None:
                self._hist.append(self._value)
                while len(self._hist) > self._depth:
                    self._hist.pop(0)
        self._value = self._fetch_value()
        return self._value

    @abstractmethod
    def _fetch_value(self) -> Any:
        return self._value


class SyscmdVariable(Variable):
    """
    A user defines a variable with its "name" and sysctl name for its value.
    """

    def __init__(self, name: str, params: Dict):
        super().__init__(name, "syscmd", params)

        self._cmd = params["syscmd"]
        self._value = self._fetch_value()

    def _fetch_value(self) -> Any:
        return os.popen(self._cmd).read().strip()


class SysctlVariable(Variable):
    """
    A user defines a variable with its "name" and sysctl name for its value.
    """

    def __init__(self, name: str, params: Dict):
        super().__init__(name, "sysctl", params)

        self._sysctl_name = params["sysctl"]
        self._sysctl = sysctl.Sysctl(self._sysctl_name)
        self._value = self._sysctl.value

    def _fetch_value(self) -> Any:
        return self._sysctl.value


def instantiate_variable(name: str, params: Dict) -> Any:
    assert name

    type = params.get("type", None)
    if type == "sysctl":
        return SysctlVariable(name, params)
    elif type == "syscmd":
        return SyscmdVariable(name, params)
    else:
        raise TypeError("Unknown variable type")
